@@ -1,28 +1,54 @@
 package autobox;
 
 use 5.008;
+
 use strict;
 use warnings;
 
-use Carp qw(confess);
+use Carp;
 use XSLoader;
-use Scalar::Util qw(blessed);
+use Scalar::Util;
 use Scope::Guard;
+use Storable;
 
-our $VERSION = '2.11';
+our $VERSION = '2.20';
 
 XSLoader::load 'autobox', $VERSION;
 
 ############################################# PRIVATE ###############################################
 
-my $SEQ   = 0;
-my $CACHE = {};    # hold a reference to the bindings hashes
+my $SEQ            = 0;  # unique identifier for synthetic classes
+my $BINDINGS_CACHE = {}; # hold a reference to the bindings hashes
+my $CLASS_CACHE    = {}; # reuse the same synthetic class if the type/superclasses are the same
 
 # create a shim class - actual methods are implemented by the classes in its @ISA
-sub _generate_class($) {
-    my $type = lc shift;
-    ++$SEQ;
-    return "autobox::$type\::<$SEQ>";
+#
+# as an optimization, or at least a courtesy, return the previously-generated class
+# if we've seen the same (canonicalized) type/isa combination before
+
+sub _generate_class($@) {
+    my ($type, $isa) = @_;
+
+    # As an optimization, simply return the class if there's only one.
+    # This speeds up method lookup as the method can (often) be found directly in the stash
+    # rather than in the ISA hierarchy with its attendant AUTOLOAD-related overhead
+
+    if (@$isa == 1) {
+        my $class = $isa->[0];
+        _universalize($class); # nop if it's already been universalized
+        return $class;
+    }
+
+    my $key = Storable::freeze([ $type, sort(@$isa) ]); # sort() - canonicalize the @isa
+
+    return $CLASS_CACHE->{$key} ||= do {
+        my $class = sprintf('autobox::shim::<%d>', ++$SEQ);
+        my $synthetic_class_isa = _get_isa($class); # i.e. autovivify
+
+        @$synthetic_class_isa = @$isa;
+        _universalize($class);
+        $class;
+    };
 }
 
 # make can() and isa() work as expected for autoboxed values by calling the method on their associated class
@@ -36,10 +62,27 @@ sub _universalize ($) {
     }
 }
 
+# pretty-print the bindings hash by showing its values as the types classes rather than the synthetic class
+sub _annotate($) {
+    my $hash = { %{ shift() } };
+
+    # reverse() turns a hash that maps a type/isa signature to a class name into a hash that maps
+    # a class name into a boolean
+    my %synthetic = reverse(%$CLASS_CACHE);
+
+    for my $key (keys %$hash) {
+        my $value = $hash->{$key};
+        $hash->{$key} = join(', ', _get_isa($value)) if ($synthetic{$value});
+    }
+
+    return $hash;
+}
+
 # default method called when the DEBUG option is supplied with a true value
 # prints the assigned bindings for the current scope along with a list of classes in their @ISA
+
 sub _debug ($) {
-    my $bindings = annotate(shift);
+    my $bindings = _annotate(shift);
     require Data::Dumper;
     no warnings qw(once);
     local ($|, $Data::Dumper::Indent, $Data::Dumper::Terse, $Data::Dumper::Sortkeys) = (1, 1, 1, 1);
@@ -49,7 +92,7 @@ sub _debug ($) {
 # return true if $ref ISA $class - works with non-references, unblessed references and objects
 sub _isa($$) {
     my ($ref, $class) = @_;
-    return blessed($ref) ? $ref->isa($class) : ref($ref) eq $class;
+    return Scalar::Util::blessed($ref) ? $ref->isa($class) : ref($ref) eq $class;
 }
 
 # get the @ISA for the specified class
@@ -63,32 +106,14 @@ sub _get_isa($) {
 }
 
 # install a new set of bindings for the current scope
+# XXX this could be refined to reuse the same hashref if its contents have already already seen,
+# but that requires each (frozen) hash to be cached; at best, it may not be much of a win, and at
+# worst it will increase bloat
+
 sub _register ($) {
     my $bindings = shift;
-
-    # As an optimization, the synthetic class is replaced with the actual class if there's
-    # only one. This speeds up method lookup as the method can (often) be found directly in the stash
-    # rather than in the ISA hierarchy with its attendant AUTOLOAD-related overhead
-
-    my $optimize = sub {
-        for my $key (keys %$bindings) {
-            my $class = $bindings->{$key};
-            my @isa   = _get_isa($class);
-            if (@isa == 1) {
-                $class = $isa[0];
-                $bindings->{$key} = $class;
-                _universalize($class);
-            }
-        }
-    };
-
-    my $sg = Scope::Guard->new($optimize);
-
     $^H{autobox} = $bindings;
-    # we can safely clobber (i.e. trigger) the previous optimizer now we've assigned
-    # a new binding for this scope
-    $^H{autobox_optimize} = $sg;
-    $CACHE->{$bindings} = $bindings; # keep the $bindings hash alive
+    $BINDINGS_CACHE->{$bindings} = $bindings; # keep the $bindings hash alive
 }
 
 ############################################# PUBLIC (Methods) ###############################################
@@ -111,17 +136,6 @@ sub defaults {
     };
 }
 
-# This undocumented method takes a bindings hash and appends a list of superclasses to its values
-# It's used by the test suite (hence the public-looking name) but is otherwise private
-sub annotate($) {
-    my $hash = { %{ shift() } };
-    for my $key (keys %$hash) {
-        my $value = $hash->{$key};
-        $hash->{$key} = join(', ', _get_isa($value));
-    }
-    return $hash;
-}
-
 # enable some flavour of autoboxing in the current scope
 sub import {
     my ($class, %args) = @_;
@@ -129,7 +143,11 @@ sub import {
     my $debug    = delete $args{DEBUG};
 
     # Don't do this until DEBUG has been deleted
-    %args = %$defaults unless (%args);
+    unless (%args) {
+        for my $key (keys %$defaults) {
+            $args{$key} = $defaults->{$key} if ($defaults->{$key});
+        }
+    }
 
     my $default = exists $args{DEFAULT} ? delete $args{DEFAULT} : $defaults->{DEFAULT};
     my $bindings; # custom typemap
@@ -159,7 +177,7 @@ sub import {
     # fill in defaults for unhandled cases; namespace expansion is handled below
     if (defined $default) {
         for my $key (keys %$defaults) {
-            next if (exists $args{$key});    # don't supply a default if the binding is already explicitly defined
+            next if (exists $args{$key});    # don't supply a default if the binding is explicitly defined
             next if ($key eq 'UNDEF');       # UNDEF must be autoboxed explicitly - DEFAULT doesn't include it
             next if ($key eq 'DEFAULT');     # already merged into $default above
 
@@ -170,35 +188,20 @@ sub import {
     # sanity check %args, expand the namespace prefixes into class names,
     # and copy defined values to the $bindings hash
 
-    for my $key (keys %args) {
-        confess("unrecognised option: '", (defined $key ? $key : '<undef>'), "'") unless (exists $defaults->{$key});
+    for my $type (keys %args) {
+        Carp::confess("unrecognized option: '", (defined $type ? $type : '<undef>'), "'")
+            unless (exists $defaults->{$type});
 
-        my $value = $args{$key};
+        my $value = $args{$type};
 
-        next unless (defined $value);
+        next unless ($value);
 
-        my $outer_class = exists($bindings->{$key}) ? $bindings->{$key} : undef; # don't autovivify
-        my ($synthetic_class, $new_synthetic_class);
-
-        if ($new_scope || not($outer_class)) {
-            # new synthetic class - either a new scope, or a new type in an existing scope
-            $synthetic_class = _generate_class($key);
-            $new_synthetic_class = 1;
-        } else {
-            $synthetic_class = $outer_class;
-            $new_synthetic_class = 0;
-        }
-
-        my $synthetic_class_isa = _get_isa($synthetic_class);
-
-        # if this is a new nested scope, merge superclasses for this type from the outer scope
-        if ($new_scope && $outer_class) {
-            my $outer_isa = _get_isa($outer_class);
-
-            for my $merge_class (@$outer_isa) {
-                push(@$synthetic_class_isa, $merge_class)
-                    unless (grep { $_ eq $merge_class } @$synthetic_class_isa); # no dups
-            }
+        my @isa = ();
+        my %synthetic = reverse (%$CLASS_CACHE); # synthetic class name => bool - see _annotate
+       
+        if (exists($bindings->{$type})) { # don't autovivify
+            my $class = $bindings->{$type};
+            @isa = $synthetic{$class} ? _get_isa($class) : ($class);
         }
 
         # we can't use UNIVERSAL::isa to test if $value is an array ref;
@@ -206,27 +209,24 @@ sub import {
 
         $value = [ $value ] unless (_isa($value, 'ARRAY'));
 
-        for my $user_class (@$value) {
-            # squashed bug: if $user_class is a default namespace that was passed in an array ref,
-            # then that array ref may have been assigned as the default for multiple types
+        for my $class (@$value) {
+            # squashed bug: if $class is a default namespace that was passed in an array ref,
+            # then that array ref may have been assigned (above) as the default for multiple types
             #
-            # mutating it (e.g. $user_class = "$user_class::$key") mutates it for all
-            # the types that have been assigned this default (due to the aliasing semantics of foreach). which
+            # mutating it (e.g. $class = "$class::$type") mutates it for all
+            # the types that default to this namespace (due to the aliasing semantics of foreach). which
             # means defaulted types all inherit MyNamespace::CODE (or whatever comes first)
             #
-            # creating a copy of $user_class and appending to that rather than mutating the original fixes this
+            # creating a copy of $class and appending to that rather than mutating the original fixes this
             # 
             # see tests 27 and 28 in merge.t
-            my $expanded = ($user_class =~ /::$/) ? "$user_class$key" : $user_class; # handle namespace expansion
+            my $expanded = ($class =~ /::$/) ? "$class$type" : $class; # handle namespace expansion
 
-            push(@$synthetic_class_isa, $expanded)
-                unless (grep { $_ eq $expanded } @$synthetic_class_isa); # no dups
+            push (@isa, $expanded) unless (grep { $_ eq $expanded } @isa); # no dups
         }
 
-        if ($new_synthetic_class) {
-            _universalize($synthetic_class);
-            $bindings->{$key} = $synthetic_class;
-        }
+        # associate the synthetic class with the specified type
+        $bindings->{$type} = _generate_class($type, \@isa);
     }
 
     # This turns on autoboxing i.e. the method call checker sets a flag on the method call op
@@ -235,14 +235,14 @@ sub import {
     # It needs to be set unconditionally because it may have been unset in unimport
     $^H |= 0x120000; # set HINT_LOCALIZE_HH + an unused bit to work around a %^H bug
 
-    # install the specified bindings in the current scope
-    _register($bindings);
-
     if ($debug) {
         $debug = \&_debug unless (_isa($debug, 'CODE'));
-        $debug->($bindings);
+        $debug->(_annotate($bindings));
     }
 
+    # install the specified bindings in the current scope
+    _register($bindings);
+
     return unless ($new_scope);
 
     # This sub is called when the current scope's %^H is destroyed i.e. at the end of the compilation of the scope
@@ -285,18 +285,19 @@ sub unimport {
     # not yet been turned on
     return unless ($^H{autobox});
 
-    my $new_bindings = { %{$^H{autobox}} }; # clone the current bindings hash
+    my $bindings = { %{$^H{autobox}} }; # clone the current bindings hash
 
     @args = keys(%$defaults) unless (@args);
 
     for my $arg (@args) {
-        confess("unrecognised option: '", (defined $arg ? $arg : '<undef>'), "'") unless (exists $defaults->{$arg});
-        delete $new_bindings->{$arg};
+        Carp::confess("unrecognised option: '", (defined $arg ? $arg : '<undef>'), "'")
+            unless (exists $defaults->{$arg});
+        delete $bindings->{$arg};
     }
 
     # unset HINT_LOCALIZE_HH + the additional bit if there are no more bindings in this scope
-    $^H &= ~0x120000 unless (%$new_bindings);
-    _register($new_bindings);
+    $^H &= ~0x120000 unless (%$bindings);
+    _register($bindings);
 }
 
 1;
